/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2009                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include<vcg/simplex/vertex/base.h>
#include<vcg/simplex/vertex/component.h>

#include <vcg/complex/used_types.h>

#include<vcg/simplex/face/base.h>
#include<vcg/simplex/face/component.h>

#include<vcg/simplex/face/topology.h>
#include<vcg/complex/trimesh/base.h>

// input output
#include<wrap/io_trimesh/import.h>
#include<wrap/io_trimesh/export.h>

// topology computation
#include<vcg/complex/trimesh/update/topology.h>

// normals
#include<vcg/complex/trimesh/update/normal.h>

using namespace vcg;
using namespace std;

class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public UsedTypes<	Use<MyVertex>   ::AsVertexType,
                                        Use<MyEdge>     ::AsEdgeType,
                                        Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public Vertex<MyUsedTypes, vertex::Coord3f, vertex::Normal3f, vertex::BitFlags  >{};
class MyFace    : public Face< MyUsedTypes, face::FFAdj,  face::VertexRef, face::BitFlags > {};
class MyEdge    : public Edge<MyUsedTypes>{};
class MyMesh    : public tri::TriMesh< vector<MyVertex>, vector<MyFace> , vector<MyEdge>  > {};


template<class MESH_TYPE>
class OrderedTriple
{
public:
  int ind[3];
  typename MESH_TYPE::FacePointer fp;

  OrderedTriple(MESH_TYPE &m, typename MESH_TYPE::FaceType &f)
  {
    for(int i=0;i<3;++i)
    {
      ind[i] = f.V(i) - &(m.vert[0]);
    }
    std::sort(ind+0, ind+3);
    fp=&f;
  }
  bool operator < (const OrderedTriple &tri) const
  {
    if(ind[0]<tri.ind[0]) return true;
    if(ind[0]>tri.ind[0]) return false;
    if(ind[1]<tri.ind[1]) return true;
    if(ind[1]>tri.ind[1]) return false;
    if(ind[2]<tri.ind[2]) return true;
    return false;
  }
  bool operator == (const OrderedTriple &tri) const
  {
   for(int i=0;i<3;++i)
     if(ind[i]!=tri.ind[i]) return false;

   return true;
  }
};

template <class MESH_TYPE>
int RemoveDoubleFace(MESH_TYPE &m)
{

  typename MESH_TYPE::FaceIterator fi;
  std::vector<OrderedTriple<MESH_TYPE> > tripleVec;

  for(fi=m.face.begin();fi!=m.face.end();++fi)
  {
    tripleVec.push_back(OrderedTriple<MESH_TYPE>(m,*fi));
  }

  std::sort(tripleVec.begin(),tripleVec.end());

  int deletedCnt=0;
  for(int i=0;i<(int)tripleVec.size()-1;i++)
  {
    if(tripleVec[i]==tripleVec[i+1])
    {
      tri::Allocator<MESH_TYPE>::DeleteFace(m,*tripleVec[i].fp);
      tri::Allocator<MESH_TYPE>::DeleteFace(m,*tripleVec[i+1].fp);
      deletedCnt++;
      i++;
    }
  }
return deletedCnt;

}

int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.ply>\n");
    return -1;
  }

  MyMesh m;

  if(tri::io::ImporterSTL<MyMesh>::Open(m,argv[1])!=0)
  {
    printf("Error reading file  %s\n",argv[1]);
    exit(0);
  }

  int deletedVert = tri::Clean<MyMesh>::RemoveDuplicateVertex(m);
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  tri::UpdateFlags<MyMesh>::FaceClearV(m);
  tri::UpdateNormals<MyMesh>::PerVertexNormalized(m);

  int deletedCnt= RemoveDoubleFace(m);

  printf("Deleted %i vert\n",deletedVert);
  printf("Final mesh has %i face and we deleted %i\n",m.fn,deletedCnt);
  tri::io::ExporterSTL<MyMesh>::Save(m,"test.stl");
  return 0;
}
